#!/usr/bin/env python

import os
import unittest
import matplotlib
matplotlib.use("agg")

from test_data import *
from test_utils import *
from test_directi import *
from test_blast import *

if __name__ == "__main__":
    unittest.main()
