r"""
Global configuration
"""

RANDOM_SEED = 0
N_JOBS = 1

_USE_GLOBAL = "__UsE_gLoBaL__"
_NAN_REPLACEMENT = "__nAn_RePlAcEmEnT__"

H5_COMPRESS_OPTS = {
    "compression": "gzip",
    "compression_opts": 7,
}
H5_TRACK_OPTS = {
    "track_times": False
}
